@@ -13,23 +13,25 @@ from urllib.parse import urlencode
 from uuid import uuid1, getnode
 
 import ujson
+from requests import Session, Response
 from tqdm import tqdm
 
 from .config.log_config import log_config
 from .config.operations import operations
 from .config.settings import *
-from .login import Session, Response
-from .utils import get_auth_headers
+from .utils import get_headers, build_query
 
 try:
     if get_ipython().__class__.__name__ == 'ZMQInteractiveShell':
         import nest_asyncio
+
         nest_asyncio.apply()
 except:
     ...
 
 if sys.platform != 'win32':
     import uvloop
+
     asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
 else:
     asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
@@ -39,6 +41,7 @@ logger = logging.getLogger(__name__)
 
 
 class Operation(Enum):
+    # tweet
     CreateTweet = auto()
     CreateScheduledTweet = auto()
     DeleteTweet = auto()
@@ -47,12 +50,29 @@ class Operation(Enum):
     UnfavoriteTweet = auto()
     CreateRetweet = auto()
     DeleteRetweet = auto()
+    # bookmark
     CreateBookmark = auto()
     DeleteBookmark = auto()
     BookmarksAllDelete = auto()
-    TweetStats = auto()
+    # topic
+    TopicFollow = auto()
+    TopicUnfollow = auto()
+    # list
+    ListsManagementPageTimeline = auto()
+    CreateList = auto()
+    DeleteList = auto()
+    EditListBanner = auto()
+    DeleteListBanner = auto()
+    ListAddMember = auto()
+    ListRemoveMember = auto()
+    ListsPinMany = auto()
+    ListPinOne = auto()
+    ListUnpinOne = auto()
+    UpdateList = auto()
     # DM
     useSendMessageMutation = auto()
+    # other
+    TweetStats = auto()
 
 
 def log(fn=None, *, level: int = logging.DEBUG, info: list = None) -> callable:
@@ -70,7 +90,7 @@ def log(fn=None, *, level: int = logging.DEBUG, info: list = None) -> callable:
         limits = {k: v for k, v in r.headers.items() if 'x-rate-limit' in k}
         current_time = int(time.time())
         wait = int(r.headers.get('x-rate-limit-reset', current_time)) - current_time
-        logger.log(level, f'{WARN}{wait // 60} minutes{RESET} until rate-limit reset. {limits = }')
+        logger.log(level, f'{wait // 60} minutes until rate-limit reset. {limits = }')
 
         try:
             if 200 <= r.status_code < 300:
@@ -100,12 +120,12 @@ def graphql_request(_id: int, operation: any, key: str | int, session: Session)
     qid = params['queryId']
     if key: params['variables'][key] = _id
     url = f"https://api.twitter.com/graphql/{qid}/{operation}"
-    r = session.post(url, headers=get_auth_headers(session), json=params)
+    r = session.post(url, headers=get_headers(session), json=params)
     return r
 
 
 def api_request(settings: dict, path: str, session: Session) -> Response:
-    headers = get_auth_headers(session)
+    headers = get_headers(session)
     headers['content-type'] = 'application/x-www-form-urlencoded'
     url = f'https://api.twitter.com/1.1/{path}'
     r = session.post(url, headers=headers, data=urlencode(settings))
@@ -120,7 +140,7 @@ def upload_media(filename: str, session: Session, is_dm: bool = False, is_profil
 
     file = Path(filename)
     total_bytes = file.stat().st_size
-    headers = get_auth_headers(session)
+    headers = get_headers(session)
 
     upload_type = 'dm' if is_dm else 'tweet'
     media_type = mimetypes.guess_type(file)[0]
@@ -136,7 +156,9 @@ def upload_media(filename: str, session: Session, is_dm: bool = False, is_profil
     data = {'command': 'INIT', 'media_type': media_type, 'total_bytes': total_bytes, 'media_category': media_category}
     r = session.post(url=url, headers=headers, data=data)
     media_id = r.json()['media_id']
-    with tqdm(total=total_bytes, desc=f"uploading: {file.name}") as pbar:
+
+    desc = f"uploading: {file.name}"
+    with tqdm(total=total_bytes, desc=desc, unit='B', unit_scale=True, unit_divisor=1024) as pbar:
         with open(file, 'rb') as f:
             i = 0
             while chunk := f.read(4 * 1024 * 1024):  # todo: arbitrary max size for now
@@ -177,7 +199,7 @@ def upload_media(filename: str, session: Session, is_dm: bool = False, is_profil
 def add_alt_text(text: str, media_id: int, session: Session) -> Response:
     params = {"media_id": media_id, "alt_text": {"text": text}}
     url = 'https://api.twitter.com/1.1/media/metadata/create.json'
-    r = session.post(url, headers=get_auth_headers(session), json=params)
+    r = session.post(url, headers=get_headers(session), json=params)
     return r
 
 
@@ -213,7 +235,7 @@ def tweet(text: str, session: Session, media: list[dict | str] = None, **kwargs)
         params['variables'] |= poll_params
 
     url = f"https://api.twitter.com/graphql/{qid}/{operation}"
-    r = session.post(url, headers=get_auth_headers(session), json=params)
+    r = session.post(url, headers=get_headers(session), json=params)
     return r
 
 
@@ -320,10 +342,99 @@ def unbookmark(_id: int, session: Session) -> Response:
 #     return graphql_request(0, Operation.BookmarksAllDelete.name, 0, session)
 
 
+@log(info=['json'])
+def stats(rest_id: int, session: Session) -> Response:
+    """private endpoint?"""
+    operation = Operation.TweetStats.name
+    params = deepcopy(operations[operation])
+    qid = params['queryId']
+    params['variables']['rest_id'] = rest_id
+    query = build_query(params)
+    url = f"https://api.twitter.com/graphql/{qid}/{operation}?{query}"
+    r = session.get(url, headers=get_headers(session))
+    return r
+
+
+@log(info=['json'])
+def dm(text: str, receivers: list[int], session: Session, filename: str = '') -> Response:
+    operation = Operation.useSendMessageMutation.name
+    params = deepcopy(operations[operation])
+    qid = params['queryId']
+    params['variables']['target'] = {"participant_ids": receivers}
+    params['variables']['requestId'] = str(uuid1(getnode()))  # can be anything
+    url = f"https://api.twitter.com/graphql/{qid}/{operation}"
+    if filename:
+        media_id = upload_media(filename, session, is_dm=True)
+        params['variables']['message']['media'] = {'id': media_id, 'text': text}
+    else:
+        params['variables']['message']['text'] = {'text': text}
+    r = session.post(url, headers=get_headers(session), json=params)
+    return r
+
+
+@log(info=['json'])
+def update_profile_image(filename: str, session: Session) -> Response:
+    media_id = upload_media(filename, session, is_profile=True)
+    url = 'https://api.twitter.com/1.1/account/update_profile_image.json'
+    headers = get_headers(session)
+    params = {'media_id': media_id}
+    r = session.post(url, headers=headers, params=params)
+    return r
+
+
+@log
+def update_profile_banner(filename: str, session: Session) -> Response:
+    media_id = upload_media(filename, session, is_profile=True)
+    url = 'https://api.twitter.com/1.1/account/update_profile_banner.json'
+    headers = get_headers(session)
+    params = {'media_id': media_id}
+    r = session.post(url, headers=headers, params=params)
+    return r
+
+
+@log
+def update_profile_info(session: Session, **kwargs) -> Response:
+    url = 'https://api.twitter.com/1.1/account/update_profile.json'
+    headers = get_headers(session)
+    r = session.post(url, headers=headers, params=kwargs)
+    return r
+
+
+@log(info=['json'])
+def create_poll(text: str, choices: list[str], poll_duration: int, session: Session) -> Response:
+    options = {
+        "twitter:card": "poll4choice_text_only",
+        "twitter:api:api:endpoint": "1",
+        "twitter:long:duration_minutes": poll_duration  # max: 10080
+    }
+    for i, c in enumerate(choices):
+        options[f"twitter:string:choice{i + 1}_label"] = c
+
+    headers = get_headers(session)
+    headers['content-type'] = 'application/x-www-form-urlencoded'
+    url = 'https://caps.twitter.com/v2/cards/create.json'
+    r = session.post(url, headers=headers, params={'card_data': ujson.dumps(options)})
+    card_uri = r.json()['card_uri']
+    r = tweet(text, session, poll_params={'card_uri': card_uri})
+    return r
+
+
+@log(info=['json'])
+def pin(tweet_id: int, session: Session) -> Response:
+    settings = {'tweet_mode': 'extended', 'id': tweet_id}
+    return api_request(settings, 'account/pin_tweet.json', session)
+
+
+@log(info=['json'])
+def unpin(tweet_id: int, session: Session) -> Response:
+    settings = {'tweet_mode': 'extended', 'id': tweet_id}
+    return api_request(settings, 'account/unpin_tweet.json', session)
+
+
 @log(info=['text'])
 def update_search_settings(session: Session, **kwargs) -> Response:
     twid = int(session.cookies.get_dict()['twid'].split('=')[-1].strip('"'))
-    headers = get_auth_headers(session=session)
+    headers = get_headers(session=session)
     r = session.post(
         url=f'https://api.twitter.com/1.1/strato/column/User/{twid}/search/searchSafety',
         headers=headers,
@@ -344,94 +455,196 @@ def update_content_settings(session: Session, **kwargs) -> Response:
     return api_request(kwargs, 'account/settings.json', session)
 
 
-def build_query(params: dict) -> str:
-    return '&'.join(f'{k}={ujson.dumps(v)}' for k, v in params.items())
+@log(info=['json'])
+def remove_interests(session: Session, *args):
+    url = 'https://api.twitter.com/1.1/account/personalization/twitter_interests.json'
+    r = session.get(url, headers=get_headers(session))
+    current_interests = r.json()['interested_in']
+    if args == 'all':
+        disabled_interests = [x['id'] for x in current_interests]
+    else:
+        disabled_interests = [x['id'] for x in current_interests if x['display_name'] in args]
+    payload = {
+        "preferences": {
+            "interest_preferences": {
+                "disabled_interests": disabled_interests,
+                "disabled_partner_interests": []
+            }
+        }
+    }
+    url = 'https://api.twitter.com/1.1/account/personalization/p13n_preferences.json'
+    r = session.post(url, headers=get_headers(session), json=payload)
+    return r
 
 
 @log(info=['json'])
-def stats(rest_id: int, session: Session) -> Response:
-    """private endpoint?"""
-    operation = Operation.TweetStats.name
+def __get_lists(session: Session) -> Response:
+    operation = Operation.ListsManagementPageTimeline.name
     params = deepcopy(operations[operation])
     qid = params['queryId']
-    params['variables']['rest_id'] = rest_id
     query = build_query(params)
     url = f"https://api.twitter.com/graphql/{qid}/{operation}?{query}"
-    r = session.get(url, headers=get_auth_headers(session))
+    r = session.get(url, headers=get_headers(session))
     return r
 
 
 @log(info=['json'])
-def dm(text: str, receivers: list[int], session: Session, filename: str = '') -> Response:
-    operation = Operation.useSendMessageMutation.name
-    params = deepcopy(operations[operation])
-    qid = params['queryId']
-    params['variables']['target'] = {"participant_ids": receivers}
-    params['variables']['requestId'] = str(uuid1(getnode()))  # can be anything
-    url = f"https://api.twitter.com/graphql/{qid}/{operation}"
-    if filename:
-        media_id = upload_media(filename, session, is_dm=True)
-        params['variables']['message']['media'] = {'id': media_id, 'text': text}
-    else:
-        params['variables']['message']['text'] = {'text': text}
-    r = session.post(url, headers=get_auth_headers(session), json=params)
-    return r
-
-
-@log(info=['json'])
-def update_profile_image(filename: str, session: Session) -> Response:
-    media_id = upload_media(filename, session, is_profile=True)
-    url = 'https://api.twitter.com/1.1/account/update_profile_image.json'
-    headers = get_auth_headers(session)
-    params = {'media_id': media_id}
-    r = session.post(url, headers=headers, params=params)
-    return r
-
-
-@log
-def update_profile_banner(filename: str, session: Session) -> Response:
-    media_id = upload_media(filename, session, is_profile=True)
-    url = 'https://api.twitter.com/1.1/account/update_profile_banner.json'
-    headers = get_auth_headers(session)
-    params = {'media_id': media_id}
-    r = session.post(url, headers=headers, params=params)
-    return r
-
-
-@log
-def update_profile_info(session: Session, **kwargs) -> Response:
-    url = 'https://api.twitter.com/1.1/account/update_profile.json'
-    headers = get_auth_headers(session)
-    r = session.post(url, headers=headers, params=kwargs)
-    return r
-
-
-@log(info=['json'])
-def create_poll(text: str, choices: list[str], poll_duration: int, session: Session) -> Response:
-    options = {
-        "twitter:card": "poll4choice_text_only",
-        "twitter:api:api:endpoint": "1",
-        "twitter:long:duration_minutes": poll_duration  # max: 10080
+def create_list(session: Session, name: str, description: str, private: bool) -> Response:
+    operation = Operation.CreateList.name
+    payload = deepcopy(operations[operation])
+    qid = payload['queryId']
+    payload['variables'] |= {
+        "isPrivate": private,
+        "name": name,
+        "description": description,
     }
-    for i, c in enumerate(choices):
-        options[f"twitter:string:choice{i + 1}_label"] = c
-
-    headers = get_auth_headers(session)
-    headers['content-type'] = 'application/x-www-form-urlencoded'
-    url = 'https://caps.twitter.com/v2/cards/create.json'
-    r = session.post(url, headers=headers, params={'card_data': ujson.dumps(options)})
-    card_uri = r.json()['card_uri']
-    r = tweet(text, session, poll_params={'card_uri': card_uri})
+    url = f"https://api.twitter.com/graphql/{qid}/{operation}"
+    r = session.post(url, headers=get_headers(session), json=payload)
     return r
 
 
 @log(info=['json'])
-def pin(tweet_id: int, session: Session) -> Response:
-    settings = {'tweet_mode': 'extended', 'id': tweet_id}
-    return api_request(settings, 'account/pin_tweet.json', session)
+def update_list(session: Session, list_id: int, name: str, description: str, private: bool) -> Response:
+    operation = Operation.UpdateList.name
+    payload = deepcopy(operations[operation])
+    qid = payload['queryId']
+    payload['variables'] |= {
+        "listId": list_id,
+        "isPrivate": private,
+        "name": name,
+        "description": description,
+    }
+    url = f"https://api.twitter.com/graphql/{qid}/{operation}"
+    r = session.post(url, headers=get_headers(session), json=payload)
+    return r
 
 
 @log(info=['json'])
-def unpin(tweet_id: int, session: Session) -> Response:
-    settings = {'tweet_mode': 'extended', 'id': tweet_id}
-    return api_request(settings, 'account/unpin_tweet.json', session)
+def update_pinned_lists(session: Session, list_ids: list[int]) -> Response:
+    """
+    Update pinned lists
+
+    Reset all pinned lists and pin all specified lists in the order they are provided.
+
+    @param session: authenticated session
+    @param list_ids: list of list ids to pin
+    @return: response
+    """
+    operation = Operation.ListsPinMany.name
+    payload = deepcopy(operations[operation])
+    qid = payload['queryId']
+    payload['variables']['listIds'] = list_ids
+    url = f"https://api.twitter.com/graphql/{qid}/{operation}"
+    r = session.post(url, headers=get_headers(session), json=payload)
+    return r
+
+
+@log(info=['json'])
+def pin_list(session: Session, list_id: int) -> Response:
+    operation = Operation.ListPinOne.name
+    payload = deepcopy(operations[operation])
+    qid = payload['queryId']
+    payload['variables']['listId'] = list_id
+    url = f"https://api.twitter.com/graphql/{qid}/{operation}"
+    r = session.post(url, headers=get_headers(session), json=payload)
+    return r
+
+
+@log(info=['json'])
+def unpin_list(session: Session, list_id: int) -> Response:
+    operation = Operation.ListUnpinOne.name
+    payload = deepcopy(operations[operation])
+    qid = payload['queryId']
+    payload['variables']['listId'] = list_id
+    url = f"https://api.twitter.com/graphql/{qid}/{operation}"
+    r = session.post(url, headers=get_headers(session), json=payload)
+    return r
+
+
+@log(info=['json'])
+def add_list_member(session: Session, list_id: int, user_id: int) -> Response:
+    operation = Operation.ListAddMember.name
+    payload = deepcopy(operations[operation])
+    qid = payload['queryId']
+    payload['variables'] |= {
+        "listId": list_id,
+        "userId": user_id,
+    }
+    url = f"https://api.twitter.com/graphql/{qid}/{operation}"
+    r = session.post(url, headers=get_headers(session), json=payload)
+    return r
+
+
+@log(info=['json'])
+def remove_list_member(session: Session, list_id: int, user_id: int) -> Response:
+    operation = Operation.ListRemoveMember.name
+    payload = deepcopy(operations[operation])
+    qid = payload['queryId']
+    payload['variables'] |= {
+        "listId": list_id,
+        "userId": user_id,
+    }
+    url = f"https://api.twitter.com/graphql/{qid}/{operation}"
+    r = session.post(url, headers=get_headers(session), json=payload)
+    return r
+
+
+@log(info=['json'])
+def delete_list(session: Session, list_id: int) -> Response:
+    operation = Operation.DeleteList.name
+    payload = deepcopy(operations[operation])
+    qid = payload['queryId']
+    payload['variables']['listId'] = list_id
+    url = f"https://api.twitter.com/graphql/{qid}/{operation}"
+    r = session.post(url, headers=get_headers(session), json=payload)
+    return r
+
+
+@log(info=['json'])
+def update_list_banner(session: Session, list_id: int, filename: str) -> Response:
+    media_id = upload_media(filename, session)
+    operation = Operation.EditListBanner.name
+    payload = deepcopy(operations[operation])
+    qid = payload['queryId']
+    payload['variables'] |= {
+        'listId': list_id,
+        'mediaId': media_id,
+    }
+    url = f"https://api.twitter.com/graphql/{qid}/{operation}"
+    r = session.post(url, headers=get_headers(session), json=payload)
+    return r
+
+
+@log(info=['json'])
+def delete_list_banner(session: Session, list_id: int) -> Response:
+    operation = Operation.DeleteListBanner.name
+    payload = deepcopy(operations[operation])
+    qid = payload['queryId']
+    payload['variables']['listId'] = list_id
+    url = f"https://api.twitter.com/graphql/{qid}/{operation}"
+    r = session.post(url, headers=get_headers(session), json=payload)
+    return r
+
+
+@log(info=['json'])
+def follow_topic(session: Session, topic_id: int) -> Response:
+    operation = Operation.TopicFollow.name
+    payload = deepcopy(operations[operation])
+    qid = payload['queryId']
+    # Topic operations are the only ones which require the id to be a string, weird
+    payload['variables']['topicId'] = str(topic_id)
+    url = f"https://api.twitter.com/graphql/{qid}/{operation}"
+    r = session.post(url, headers=get_headers(session), json=payload)
+    return r
+
+
+@log(info=['json'])
+def unfollow_topic(session: Session, topic_id: int) -> Response:
+    operation = Operation.TopicUnfollow.name
+    payload = deepcopy(operations[operation])
+    qid = payload['queryId']
+    # Topic operations are the only ones which require the id to be a string, weird
+    payload['variables']['topicId'] = str(topic_id)
+    url = f"https://api.twitter.com/graphql/{qid}/{operation}"
+    r = session.post(url, headers=get_headers(session), json=payload)
+    return r
